
/* mkwords.c v0.1a (c) Ilya V. Vasilyev, Moscow, 2016 */
/* Часть проекта Каллисто                             */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXWORDS (330)
#define NCERS (13)

struct TCER
{ const char *tag;		/* Как CER выглядит в mkl          */
  const char *u;		/* Как CER выглядит в Unicode      */
  const unsigned char mk;	/* Как CER выглядит в кодировке МК */
};

static struct TCER sCERs [NCERS] =	/* Все CER'ы, известные mkwords */
{ { "&amp", "&", 38 },
  { "&times", "×", 179 },
  { "&divide", "÷", 176 },
  { "&crarr", "↵", 192 },
  { "&#8598", "↖", 200 },
  { "&#8744", "∨", 118 },
  { "&#8853", "⊕", 181 },
  { "&#739", "ˣ", 190 },
  { "&sup2", "²", 189 },
  { "&radic", "√", 251 },
  { "&harr", "↔", 183 },
  { "&uarr", "↑", 24 },
  { "&pi", "π", 185 }
};

static const char *cp1251[256] =		/* Перевод из CP1251 в Юникод */
{ "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
  " ", "!", "\"", "#", "$", "%", "&", "'", "(", ")", "∗", "+", ",", "−", ".", "/",
  "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", ":", ";", "<", "=", ">", "?",
  "@", "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O",
  "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z", "[", "\\", "]", "^", "_",
  "`", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o",
  "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "{", "|", "}", "∼", "",
  "Ђ", "Ѓ", "‚", "ѓ", "„", "…", "†", "‡", "€", "‰", "Љ", "‹", "Њ", "Ќ", "Ћ", "Џ",
  "ђ", "‘", "’", "“", "”", "•", "–", "—", "", "™", "љ", "›", "њ", "ќ", "ћ", "џ",
  " ", "Ў", "ў", "Ј", "¤", "Ґ", "¦", "§", "Ё", "©", "Є", "«", "¬", "-", "®", "Ї",
  "°", "±", "І", "і", "ґ", "µ", "¶", "·", "ё", "№", "є", "»", "ј", "Ѕ", "ѕ", "ї",
  "А", "Б", "В", "Г", "Д", "Е", "Ж", "З", "И", "Й", "К", "Л", "М", "Н", "О", "П",
  "Р", "С", "Т", "У", "Ф", "Х", "Ц", "Ч", "Ш", "Щ", "Ъ", "Ы", "Ь", "Э", "Ю", "Я",
  "а", "б", "в", "г", "д", "е", "ж", "з", "и", "й", "к", "л", "м", "н", "о", "п",
  "р", "с", "т", "у", "ф", "х", "ц", "ч", "ш", "щ", "ъ", "ы", "ь", "э", "ю", "я"
};

static const unsigned char mkcharset[256] =	/* Перевод из CP1251 в кодировку МК-161 */
{ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
  0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
  0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, /* 0…? */
  0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, /* @…O */
  0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, /* P…_ */
  0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, /* `…o */
  0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, /* p…7f */

    72,  131,   44,  163,   34,   46,   43,  178,   36,   37,  139,   60,  141,  138,   72,   85,
   104,   96,   39,   34,   34,   42,   45,   45,   32,  116,  171,   62,  173,  170,  104,  117,
    32,  147,  227,   74,   36,  131,  124,  185,  240,   99,  157,   34,  195,   45,   82,   73,
   111,   43,   73,  105,  163,  117,  143,   42,  241,   78,  237,   34,  106,   83,  115,  105,

  0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, /* А…П */
  0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, /* Р…Я */
  0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, /* а…п */
  0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, /* р…я */
};

static unsigned char line[512], lineUnicode[2100];	/* Строка из mkl и она же в Unicode */
static char *wordlines[MAXWORDS];			/* Глоссарий в Unicode              */
static unsigned char words[MAXWORDS][33];		/* Имена слов в кодировке МК-161    */
static int sortw[MAXWORDS];				/* Отсортированные слова, их номера */

static int nWords=0;		/* Количество найденных слов           */
static int pUnicode;		/* Указатель внутри строки lineUnicode */

void WordsFree(void)		/* Освободить занятую память */
{ int i=0;
  while (i<nWords)
  { free(wordlines[i]);
    wordlines[i++] = NULL;
  }
}

void Abort(const char s[])			/* Аварийный выход в ОС */
{ fprintf(stderr, "%s\n", s);
  WordsFree();
  exit(EXIT_FAILURE);
}

void UnicodeAdd(const char *s)			/* Добавить литер к lineUnicode */
{ int i=0;
  while (s[i]!='\0')
    lineUnicode[pUnicode++] = s[i++];
}

unsigned char UnicodeAddCER(const char *s)	/* Найти CER и добавить его к lineUnicode */
{ int i=0;
  while(i<NCERS)
  { if (strcmp(s,sCERs[i].tag)==0)
    { UnicodeAdd(sCERs[i].u);
      return sCERs[i].mk;
    }
    ++i;
  }
  fprintf(stderr,"%s","Неизвестный CER: ");
  Abort(s);
  return ' '; /* Чтобы заткнуть компилятор. */
}

void WordsAdd(int ln)				/* Добавить в Глоссарий строчку line с номером ln */
{ int x=2, i=2; /* Нам не нужны ; и пробел в начале каждой строки */
  char *s1;
  char s[10], f;
  unsigned char c, cn;
  unsigned char name[33];
  int in;

  if (nWords>=MAXWORDS) Abort("Слишком много слов.");

  /* Номер строки запишем по простому, кодировка цифр совпадает */
  pUnicode=0;
  sprintf(s, "%4i ", ln);
  UnicodeAdd(s);

  /* Преобразовать строчку line в Unicode     */
  f = 1; in = 0; /* Происходит паржинг имени */
  while((c=line[i])!='\0')
  { if (c=='&')
    { int j=0; /* Считываем CER */
      do { s[j++] = c;
           c = line[++i];
         } while (c!=';');
      s[j++] = '\0';
      cn = UnicodeAddCER(s);
      if (f!=0) name[in++]=cn;
      ++x;
    }
    else if (c!=9)
    { UnicodeAdd(cp1251[c]);
      if (f!=0)
      { if (c<=' ') { name[in++]='\0'; f=0; }
        else { name[in++]=mkcharset[c]; if (in>30) Abort("Слишком длинное имя."); }
      }
      ++x;
    }
    else
    { /* Обработка табуляций */
      if (f!=0) { name[in++]='\0'; f=0; }
      do { UnicodeAdd(cp1251[32]); ++x; }
      while ((x&7)!=0);
    }
    ++i;
  }
  lineUnicode[pUnicode++] = '\0';

  /* Добавить строку в Юникоде к Глоссарию */
  s1 = malloc(pUnicode);
  if (s1==NULL) Abort("Ошибка выделения памяти.");
  sortw[nWords] = nWords;
  memcpy(s1, lineUnicode, pUnicode);
  memcpy(words[nWords], name, in);
  wordlines[nWords++] = s1;

}

int compword(const void *p1, const void *p2)	/* Сравнить два слова, для сортировки */
{ int i1 = *(int *) p1;
  int i2 = *(int *) p2;
  return strcmp((const char *)words[i1],(const char *)words[i2]);
}

void WordsSort(void)				 /* Сортируем слова по words[] */
{ qsort(sortw,nWords,sizeof(int),compword);
}

void WordsPrint(void)				/* Вывод глоссария */
{ int i=0;
  while (i<nWords)
    puts(wordlines[sortw[i++]]);
}

int main(int arc, char *argv[])
{ FILE *f = fopen(argv[1],"rb");
  int ln=1;
  fprintf(stderr, "mkwords v0.1a (c) AtH, 9 июля 2016, Файл: %s\n",argv[1]);
  if (f!=NULL)
  { puts("\nГлоссарий Каллисто v0.9\n\n");
    while(fgets((char *)line, 512, f)!=NULL)
    { if ((line[0]==';') && (line[1]=='#'))
      { fgets((char *)line, 512, f);
        ++ln; /* Нас интересует строчка сразу после метки */
        WordsAdd(ln);
      }
      ++ln;
    }
    fclose(f);
    WordsSort();
    WordsPrint();
    WordsFree();
    puts("\nЛегенда:\n"

"a        адрес в единой адресации (от 0 до 19'999)\n"
"b        8-битное целое (от -128 до 255)\n"
"c        неотрицательное 8-битное целое (от 0 до 255)\n"
"d        16-битное целое (от -32768 до 65535)\n"
"D        16-битное целое со знаком (от -32768 до 32767)\n"
"f        флаг: −1 «истина», 0 «ложь»\n"
"ф        флаг: 0 «ложь», отлично от нуля — «истина»\n"
"i        32-битное целое в дополнительном коде (от -2'147'483'648 до 2'147'483'647)\n"
"n        целое число (от -999'999'999'999 до 999'999'999'999)\n"
"p        неотрицательное десятичное число (от 0 до 9,999'999'999'99e99)\n"
"т        токен (CFA) слова\n"
"u        неотрицательное целое (от 0 до 999'999'999'999)\n"
"U        16-битное беззнаковое (от 0 до 65535)\n"
"x y z t  любой элемент стека (десятичное число от -9,999'999'999'99e99 до 9,999'999'999'99e99)\n");

  } else fprintf(stderr,
  "Не найден файл '%s'.\n"
  "Использование: mkwords имя-файла.mkl\n", argv[1]);
}


